@@ -43,7 +43,7 @@
 #endif
 
 #include <string.h>
-
+#include <unistd.h>
 #include "vesa.h"
 
 /* All drivers initialising the SW cursor need this */
@@ -66,7 +66,11 @@
 /* Mandatory functions */
 static const OptionInfoRec * VESAAvailableOptions(int chipid, int busid);
 static void VESAIdentify(int flags);
+#if defined(XSERVER_LIBPCIACCESS) && !defined(HAVE_ISA)
+#define VESAProbe NULL
+#else
 static Bool VESAProbe(DriverPtr drv, int flags);
+#endif
 #ifdef XSERVER_LIBPCIACCESS
 static Bool VESAPciProbe(DriverPtr drv, int entity_num,
      struct pci_device *dev, intptr_t match_data);
@@ -446,7 +450,15 @@ VESAPciProbe(DriverPtr drv, int entity_num, struct pci_device *dev,
 	     intptr_t match_data)
 {
     ScrnInfoPtr pScrn;
-    
+
+#ifdef __linux__
+    if (access("/sys/devices/platform/efi-framebuffer.0", F_OK) == 0 ||
+        access("/sys/devices/platform/efifb.0", F_OK) == 0) {
+        ErrorF("vesa: Refusing to run on UEFI\n");
+        return FALSE;
+    }
+#endif
+
     pScrn = xf86ConfigPciEntity(NULL, 0, entity_num, NULL, 
 				NULL, NULL, NULL, NULL, NULL);
     if (pScrn != NULL) {
@@ -466,6 +478,7 @@ VESAPciProbe(DriverPtr drv, int entity_num, struct pci_device *dev,
 }
 #endif
 
+#ifndef VESAProbe
 static Bool
 VESAProbe(DriverPtr drv, int flags)
 {
@@ -536,6 +549,7 @@ VESAProbe(DriverPtr drv, int flags)
 
     return (foundScreen);
 }
+#endif
 
 #ifdef HAVE_ISA
 static int
@@ -590,7 +604,8 @@ VESAFreeRec(ScrnInfoPtr pScrn)
     }
 #endif
     free(pVesa->monitor);
-    free(pVesa->vbeInfo);
+    if (pVesa->vbeInfo)
+	VBEFreeVBEInfo(pVesa->vbeInfo);
     free(pVesa->pal);
     free(pVesa->savedPal);
     free(pVesa->fonts);
@@ -676,12 +691,14 @@ VESAPreInit(ScrnInfoPtr pScrn, int flags)
 				    V_MODETYPE_VBE);
 
     /* Preferred order for default depth selection. */
-    if (depths & V_DEPTH_24)
+    if (depths & V_DEPTH_24 && (flags24 & Support32bppFb))
 	defaultDepth = 24;
     else if (depths & V_DEPTH_16)
 	defaultDepth = 16;
     else if (depths & V_DEPTH_15)
 	defaultDepth = 15;
+    else if (depths & V_DEPTH_24)
+        defaultDepth = 24; /* ew though */
     else if (depths & V_DEPTH_8)
 	defaultDepth = 8;
     else if (depths & V_DEPTH_4)
@@ -717,6 +734,9 @@ VESAPreInit(ScrnInfoPtr pScrn, int flags)
 
     xf86SetGamma(pScrn, gzeros);
 
+    /* set up options before loading any modules that may look at them */
+    xf86CollectOptions(pScrn, NULL);
+
     if (pVesa->major >= 2) {
 	/* Load ddc module */
 	if ((pDDCModule = xf86LoadSubModule(pScrn, "ddc")) == NULL) {
@@ -827,7 +847,6 @@ VESAPreInit(ScrnInfoPtr pScrn, int flags)
     }
 
     /* options */
-    xf86CollectOptions(pScrn, NULL);
     if (!(pVesa->Options = malloc(sizeof(VESAOptions)))) {
         vbeFree(pVesa->pVbe);
 	return FALSE;
@@ -835,12 +854,43 @@ VESAPreInit(ScrnInfoPtr pScrn, int flags)
     memcpy(pVesa->Options, VESAOptions, sizeof(VESAOptions));
     xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, pVesa->Options);
 
-    /* Use shadow by default */
-    pVesa->shadowFB = xf86ReturnOptValBool(pVesa->Options, OPTION_SHADOW_FB,
-                                           TRUE);
+    /* Use shadow by default, for non-virt hardware */
+    if (!xf86GetOptValBool(pVesa->Options, OPTION_SHADOW_FB, &pVesa->shadowFB))
+    {
+	switch (pVesa->pciInfo->vendor_id) {
+	    case 0x1234: /* bochs vga (not in pci.ids) */
+	    case 0x15ad: /* vmware */
+	    case 0x1b36: /* qemu qxl */
+	    case 0x80ee: /* virtualbox */
+	    case 0xaaaa: /* parallels (not in pci.ids) */
+		pVesa->shadowFB = FALSE;
+		break;
 
-    if (xf86ReturnOptValBool(pVesa->Options, OPTION_DFLT_REFRESH, FALSE))
-	pVesa->defaultRefresh = TRUE;
+	    case 0x1013: /* qemu's cirrus emulation */
+		if (pVesa->pciInfo->subvendor_id == 0x1af4)
+		    pVesa->shadowFB = FALSE;
+		else
+		    pVesa->shadowFB = TRUE;
+		break;
+
+	    case 0x1414: /* microsoft hyper-v */
+		if (pVesa->pciInfo->device_id == 0x5353)
+		    pVesa->shadowFB = FALSE;
+		else
+		    pVesa->shadowFB = TRUE;
+		break;
+
+	    default:
+		pVesa->shadowFB = TRUE;
+		break;
+	}
+    }
+
+    /*  Use default refresh by default. Too many VBE 3.0
+     *   BIOSes are incorrectly implemented.
+     */
+    pVesa->defaultRefresh = xf86ReturnOptValBool(pVesa->Options,
+                                                 OPTION_DFLT_REFRESH, TRUE);
 
     pVesa->ModeSetClearScreen =
         xf86ReturnOptValBool(pVesa->Options,
@@ -863,14 +913,14 @@ VESAPreInit(ScrnInfoPtr pScrn, int flags)
 		    break;
 		default:
 		    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
-			       "Unsupported bpp: %d", pScrn->bitsPerPixel);
+			       "Unsupported bpp: %d\n", pScrn->bitsPerPixel);
 		    vbeFree(pVesa->pVbe);
 		    return FALSE;
 	    }
 	    break;
 	default:
 	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
-		       "Unsupported Memory Model: %d", mode->MemoryModel);
+		       "Unsupported Memory Model: %d\n", mode->MemoryModel);
 	    return FALSE;
     }
 
@@ -1304,9 +1354,10 @@ VESAMapVidMem(ScrnInfoPtr pScrn)
 #endif
 
     xf86DrvMsgVerb(pScrn->scrnIndex, X_INFO, DEBUG_VERB,
-		   "virtual address = %p,\n"
+		   "virtual address = %p, VGAbase = %p\n"
 		   "\tphysical address = 0x%lx, size = %ld\n",
-		   pVesa->base, pScrn->memPhysBase, pVesa->mapSize);
+		   pVesa->base, pVesa->VGAbase,
+		   pScrn->memPhysBase, pVesa->mapSize);
 
     return (pVesa->base != NULL);
 }
@@ -1347,6 +1398,27 @@ VESALoadPalette(ScrnInfoPtr pScrn, int numColors, int *indices,
     int i, idx;
     int base;
 
+    if (!pVesa->savedPal) {
+#define VESADACDelay()							       \
+	do {                                                                   \
+	   (void)inb(pVesa->ioBase + VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET); \
+	   (void)inb(pVesa->ioBase + VGA_IOBASE_COLOR + VGA_IN_STAT_1_OFFSET); \
+	} while (0)
+
+	for (i = 0; i < numColors; i++) {
+	   idx = indices[i];
+	   outb(pVesa->ioBase + VGA_DAC_WRITE_ADDR, idx);
+	   VESADACDelay();
+	   outb(pVesa->ioBase + VGA_DAC_DATA, colors[idx].red);
+	   VESADACDelay();
+	   outb(pVesa->ioBase + VGA_DAC_DATA, colors[idx].green);
+	   VESADACDelay();
+	   outb(pVesa->ioBase + VGA_DAC_DATA, colors[idx].blue);
+	   VESADACDelay();
+	}
+	return;
+    }
+
     if (pVesa->pal == NULL)
 	pVesa->pal = calloc(1, sizeof(CARD32) * 256);
 
